#include "ELF.h"

#include <ehs/EHS.h>
#include <ehs/io/Console.h>
#include <ehs/io/File.h>

ehs::SInt_32 Main(ehs::Str_8* appName, ehs::Str_8* appVerId, ehs::Version* appVer)
{
	*appName = "Compiler";
	*appVerId = "Alpha";
	*appVer = {1, 0, 0};

	ehs::Vector<ehs::Str_8> args = ehs::Console::GetArgs_8();
	if (args.Size() <= 1)
	{
		ehs::Console::Write_8(*appName + " " + *appVerId + " v" + ehs::Str_8::FromNum(appVer->major) + "." +
			ehs::Str_8::FromNum(appVer->minor) + "." + ehs::Str_8::FromNum(appVer->patch));

		return 0;
	}

	ehs::File codeFile(args[1].RemoveAll("\""), ehs::Mode::READ, ehs::Disposition::OPEN);
	ehs::Serializer<ehs::UInt_64> code = codeFile.ReadSerializer_64(ehs::Endianness::LE, codeFile.Size());
	codeFile.Release();

	char sections[] = "\0.text\0.shstrtab\0";

	ElfHeader elfHeader = {};
	elfHeader.architecture = ELFH_ARCH_64;
	elfHeader.endianness = ELFH_END_LE;
	elfHeader.elfV1 = 1;
	elfHeader.targetABI = ELFH_ABI_SYSTEMV;
	elfHeader.abiV = 0;
	elfHeader.elfType = ELFH_TYPE_EXEC;
	elfHeader.machineArch = ELFH_MARCH_AMD_X86_64;
	elfHeader.elfV2 = 1;
	elfHeader.entryAddress = 0x400000 + ELFH_SIZE + PRGH_SIZE;
	elfHeader.programsOffset = ELFH_SIZE;
	elfHeader.sectionsOffset = ELFH_SIZE + PRGH_SIZE + code.Size() + sizeof(sections);
	elfHeader.flags = 0;
	elfHeader.headerSize = ELFH_SIZE;
	elfHeader.pHeaderSize = PRGH_SIZE;
	elfHeader.pEntries = 1;
	elfHeader.sSize = SECH_SIZE;
	elfHeader.sEntries = 3;
	elfHeader.strTabIndex = 2;

	ElfPrgHeader programHeader = {};
	programHeader.type = PRGH_TYPE_LOAD;
	programHeader.flags = PRGH_FLAG_EXEC | PRGH_FLAG_READ;
	programHeader.offset = 0;
	programHeader.vAddr = 0x400000;
	programHeader.pAddr = 0x400000;
	programHeader.fileSize = ELFH_SIZE + PRGH_SIZE + code.Size();
	programHeader.memSize = ELFH_SIZE + PRGH_SIZE + code.Size();
	programHeader.align = 0x200000;

	ElfSecHeader null = {};
	null.index = 0;
	null.type = SECH_TYPE_NULL;
	null.flags = 0;
	null.vAddr = 0;
	null.offset = 0;
	null.size = 0;
	null.associatedIndex = 0;
	null.info = 0;
	null.align = 0;
	null.entrySize = 0;

	ElfSecHeader text = {};
	text.index = 1;
	text.type = SECH_TYPE_PROGBITS;
	text.flags = SECH_FLAG_ALLOC | SECH_FLAG_EXEC;
	text.vAddr = 0x400000 + ELFH_SIZE + PRGH_SIZE;
	text.offset = ELFH_SIZE + PRGH_SIZE;
	text.size = sizeof(code.Size());
	text.associatedIndex = 0;
	text.info = 0;
	text.align = 16;
	text.entrySize = 0;

	ElfSecHeader strTbl = {};
	strTbl.index = 7;
	strTbl.type = SECH_TYPE_STRTAB;
	strTbl.flags = SECH_FLAG_STRINGS;
	strTbl.vAddr = 0;
	strTbl.offset = ELFH_SIZE + PRGH_SIZE + code.Size();
	strTbl.size = sizeof(sections);
	strTbl.associatedIndex = 0;
	strTbl.info = 0;
	strTbl.align = 0;
	strTbl.entrySize = 0;

	ehs::Serializer<ehs::UInt_64> data(ehs::Endianness::LE);
	WriteHeader(elfHeader, data);
	WriteHeader(programHeader, data);

	data.Resize(data.Size() + code.Size());
	ehs::Util::Copy(&data[data.GetOffset()], code, code.Size());
	data.SetOffset(data.GetOffset() + code.Size());

	data.Resize(data.Size() + sizeof(sections));
	ehs::Util::Copy(&data[data.GetOffset()], sections, sizeof(sections));
	data.SetOffset(data.GetOffset() + sizeof(sections));

	WriteHeader(null, data);
	WriteHeader(text, data);
	WriteHeader(strTbl, data);

	ehs::File file(args[2].RemoveAll("\""), ehs::Mode::WRITE, ehs::Disposition::CREATE_PERSISTENT);
	file.WriteSerializer_64(data);

	return 0;
}
